// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

const MODULE_NAME = "RSSMedia";
const RSSMEDIA_TITLE = "RSS Media Service";
const RSSMEDIA_FAVICON = "chrome://browser/skin/feeds/feedIcon16.png";
const CLASS_ID = Components.ID("{e83c52ca-40f9-401d-8d4f-cd244a357dcf}");
const CONTRACT_ID = "@flock.com/?photo-rss-media;1";
const SHORTNAME = "mediarss";
const TITLE = "RSS Media";

const MEDIARSS_NS = "http://search.yahoo.com/mrss/";

CU.import("resource:///modules/FlockXPCOMUtils.jsm");
CU.import("resource:///modules/FlockSvcUtils.jsm");
CU.import("resource:///modules/FlockPrefsUtils.jsm");

function loadLibraryFromSpec(aSpec)
{
  var loader = CC["@mozilla.org/moz/jssubscript-loader;1"]
               .getService(CI.mozIJSSubScriptLoader);
  loader.loadSubScript(aSpec);
}

/*************************************************************************
 * Component: flockRSSMedia
 *************************************************************************/
function flockRSSMedia() {
  loadLibraryFromSpec("chrome://flock/content/photo/photoAPI.js");
  FlockSvcUtils.nsIObserver.addDefaultMethod(this, "observe");
  FlockSvcUtils.flockIWebService.addDefaultMethod(this, "getStringBundle");
  FlockSvcUtils.flockIWebService.addDefaultMethod(this, "isHidden");

  FlockSvcUtils.flockIMediaWebService.addDefaultMethod(this, "getChannel");
  FlockSvcUtils.flockIMediaWebService.addDefaultMethod(this, "enumerateChannels");
  FlockSvcUtils.flockIMediaWebService.addDefaultMethod(this, "getMediaItemFormatter");

  this._init();
}

/*************************************************************************
 * flockRSSMedia: XPCOM Component Creation
 *************************************************************************/

flockRSSMedia.prototype = new FlockXPCOMUtils.genericComponent(
  RSSMEDIA_TITLE,
  CLASS_ID,
  CONTRACT_ID,
  flockRSSMedia,
  CI.nsIClassInfo.SINGLETON,
  [
    CI.nsIObserver,
    CI.flockIWebService,
    CI.flockIMediaWebService
  ]
);

// FlockXPCOMUtils.genericModule() categories
flockRSSMedia.prototype._xpcom_categories = [
  { category: "flockMediaProvider", entry: SHORTNAME }
];


flockRSSMedia.prototype._init =
function svs__init() {
  FlockSvcUtils.getLogger(this);
  this._logger.debug(".init()");

  // Used by getChannel() and enumerateChannels(), defined in FlockSvcUtils:
  // channels for the MediaBar
  this._channels = {
  };

  FlockSvcUtils.getCoop(this);

  this._logger.debug("initialized");
}

flockRSSMedia.prototype.shortName = SHORTNAME;
flockRSSMedia.prototype.title = TITLE;
flockRSSMedia.prototype.icon = RSSMEDIA_FAVICON;
flockRSSMedia.prototype.count = 1;

/*************************************************************************
 * flockRSSMedia: Private Data and Functions
 *************************************************************************/

flockRSSMedia.prototype._getFeed =
function flockRSSMedia_getFeed(aURL, aFlockListener) {
  var inst = this;
  this._req = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
              .createInstance(CI.nsIXMLHttpRequest);
  this._req.onreadystatechange = function fRSSM_onreadystatechange (aEvt) {
    if(inst._req.readyState == 4) {
      try {
        switch (inst._req.status) {
          case 200: 
          case 201:
          case 205:
            aFlockListener.onSuccess(inst._req.responseXML, "success");
            break;
          case 500: // Server internal error
            var err = CC["@flock.com/error;1"].createInstance(CI.flockIError);
            err.serviceErrorCode = 500;
            err.serviceErrorString = inst._req.responseText;
            aFlockListener.onError(err, null);
            break;
          default: // The server should not return any other code
            break;
        }
      } catch(e) {
        dump(e);
      }
    }
  };
  this._req.open("GET", aURL, true);
  this._req.overrideMimeType("text/xml");
  this._req.send(null);
}

flockRSSMedia.prototype.findByUsername =
function flockRSSMedia_findByUsername(aFlockListener, aUsername) {
  aFlockListener.onError(null, null);
}

/*************************************************************************
 * flockRSSMedia: flockIMediaWebService
 *************************************************************************/

flockRSSMedia.prototype.search =
function flockRSSMedia_search(aFlockListener,
                              aQueryString,
                              aCount,
                              aPage,
                              aRequestId)
{
  this._logger.debug("Attempting to run search query: " + aQueryString);
  if (aPage > 1 ) {
    // Pagination is not supported
    return;
  }

  var url = decodeURIComponent(aQueryString.split(":")[1]);

  // Workaround for Yahoo!image: they advertise an adult feed even when
  // their "safeSearch" is on!
  var show_adult = FlockPrefsUtils.getBoolPref("flock.photo.show_adult");
  if (!show_adult && url.match(/yahoo\.com/)) {
    url = url.replace(/\&adult_ok=1/, "");
  }

  function createEnum( array ) {
    return {
      QueryInterface: function cE_QueryInterface(iid) {
        if (iid.equals(Components.interfaces.nsISimpleEnumerator)) {
          return this;
        }
        throw Components.results.NS_ERROR_NO_INTERFACE;
      },
      hasMoreElements: function cE_hasMoreElements() {
        return (array.length > 0);
      },
      getNext: function cE_getNext() {
        return array.shift();
      }
    }
  }

  var inst = this;
  var listener = {
    onSuccess: function listener_onSuccess(aXml, aStatus) {
      var show_adult = FlockPrefsUtils.getBoolPref("flock.photo.show_adult");
      var sbs = CC["@mozilla.org/intl/stringbundle;1"]
                .getService(CI.nsIStringBundleService);
      var bundleUrl = "chrome://flock/locale/mediaItem/mediabar.properties";
      var bundle = sbs.createBundle(bundleUrl);

      var mediaItems = [];
      var channel = aXml.getElementsByTagName("channel")[0];
      var mediaItemList = channel.getElementsByTagName("item");
      for (i = 0; i < mediaItemList.length; i++) {
        try {
          var mediaItem = mediaItemList[i];
          if (mediaItem.getElementsByTagName("adult")[0]) {
            var adult = mediaItem.getElementsByTagName("adult")[0].firstChild.nodeValue;
            if (!show_adult && adult == "true") {
              continue;
            }
          }
          var title;
          if (mediaItem.getElementsByTagName("title")[0].firstChild) {
            title = mediaItem.getElementsByTagName("title")[0].firstChild.nodeValue;
          } else {
            title = bundle.GetStringFromName("flock.media.tooltip.untitled");
          }
          var thumb_url = null;
          var thumbTags = mediaItem.getElementsByTagNameNS(MEDIARSS_NS,
                                                           "thumbnail");
          if (thumbTags && thumbTags[0]) {
            thumb_url = thumbTags[0].getAttribute("url");
          }
          var link = mediaItem.getElementsByTagName("link")[0]
                              .firstChild
                              .nodeValue;
          var id = link.split("/").pop();
          var is_video = false;
          var media_url = null;
          if (mediaItem.getElementsByTagNameNS(MEDIARSS_NS, "content")) {
            var content = mediaItem.getElementsByTagNameNS(MEDIARSS_NS,
                                                           "content")[0];
            if (content && ((content.getAttribute("medium") == "video") ||
                            (/^video/.test(content.getAttribute("type")))))
            {
              is_video = true;
            }
            if (content && content.getAttribute("url")) {
              media_url = content.getAttribute("url");
            }
          }
          if (!media_url) {
            var enclosure = mediaItem.getElementsByTagName("enclosure");
            if (enclosure && enclosure[0]) {
              media_url = enclosure[0].getAttribute("url");
            }
          }
          if ((!is_video) && mediaItem.getElementsByTagName("group")[0]) {
            var contents = mediaItem.getElementsByTagName("group")[0]
                                    .getElementsByTagName("content");
            for (var j = 0; j < contents.length; j++) {
              var content = contents[j];
              if ((content.getAttribute("medium") == "video") ||
                  (/^video/.test(content.getAttribute("type"))))
              {
                is_video = true;
              }
            }
          }
          var newMediaItem = CC["@flock.com/photo;1"]
                             .createInstance(CI.flockIMediaItem);
          newMediaItem.init(inst.shortName, null);
          newMediaItem.webPageUrl = link;
          newMediaItem.thumbnail = thumb_url ? thumb_url : media_url;
          newMediaItem.midSizePhoto = media_url ? media_url : thumb_url;
          newMediaItem.largeSizePhoto = media_url ? media_url : thumb_url;
          newMediaItem.is_public = true;
          newMediaItem.is_video = is_video;
          newMediaItem.title = title;
          newMediaItem.id = id;
          mediaItems.push(newMediaItem);
        } catch(ex) {
          // Some attributes were missing in this image, skip it
          inst._logger.error(ex);
        }
      }

      aFlockListener.onSuccess(createEnum(mediaItems), aRequestId);
    },
    onError: aFlockListener.onError
  };

  this._getFeed(url, listener);
}

flockRSSMedia.prototype.supportsSearch =
function flockRSSMedia_supportsSearch( aQueryString ) {
  return false;
}


// readonly attribute boolean supportsUsers;
flockRSSMedia.prototype.supportsUsers = false;

flockRSSMedia.prototype.decorateForMedia =
function flockRSSMedia_decorateForMedia(aDocument) {
  DEBUG("{flockIMediaWebService}.decorateForMedia(aDocument)");
  // Discovery not supported yet
}

flockRSSMedia.prototype.getIconForQuery =
function flockRSSMedia_getIconForQuery(aQuery, aListener) {
  // Most service just return their favicon here... But we want to get the
  // icon from the site that hosts the feed.

  var url = decodeURIComponent(aQuery.split(":")[1]);
  dump("Looking at: {"+url+"}\n");
  var ios = CC["@mozilla.org/network/io-service;1"]
            .getService(CI.nsIIOService);
  var uri = ios.newURI(url, null, null);

  return uri.resolve("/favicon.ico");
}

/*************************************************************************
 * XPCOM Support - Module Construction
 *************************************************************************/

// Create array of components.
var componentsArray = [flockRSSMedia];

// Generate a module for XPCOM to find.
var NSGetModule = FlockXPCOMUtils.generateNSGetModule(MODULE_NAME,
                                                      componentsArray);
