// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

const YAHOO_LIVE_SEARCH_NAME = "Yahoo";
const YAHOO_LIVE_SEARCH_CLASSID =
  Components.ID("{88a8b490-d044-4a8b-bb03-1798864ffabf}");

const YAHOO_SEARCH_API_URL = "http://api.search.yahoo.com/WebSearchService/V1/";


const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

CU.import("resource:///modules/FlockLiveSearchProvider.jsm");

function YahooLiveSearch() {
  this._init();
}

YahooLiveSearch.prototype = FlockLiveSearchProvider.generateComponent(
  YAHOO_LIVE_SEARCH_NAME,
  YAHOO_LIVE_SEARCH_CLASSID,
  YahooLiveSearch,
  "chrome://flock/content/search/yahooIcon.png"
);

YahooLiveSearch.prototype.__defineGetter__("fullResultsUrl",
  function YLS_getter_getFullResultsUrl() {
    var prefs = CC["@mozilla.org/preferences-service;1"]
                .getService(CI.nsIPrefBranch);
    var url;
    try {
      url = prefs.getComplexValue("flock.search.yahoo.showall",
                                  CI.nsIPrefLocalizedString)
                 .data;
    } catch (ex) {
      url = prefs.getCharPref("flock.search.yahoo.showall");
    }
    return url;
  }
);

YahooLiveSearch.prototype._getQueryURL =
function YLS__getQueryURL(aQuery, aNumResults) {
  var prefs = CC["@mozilla.org/preferences-service;1"]
              .getService(CI.nsIPrefBranch);
  region = prefs.getComplexValue("flock.yahoo.search.param.region",
                                 CI.nsIPrefLocalizedString).data;

  return YAHOO_SEARCH_API_URL + "webSearch?"
         + "appid=flock-search"
         + "&query=" + encodeURIComponent(aQuery)
         + "&zip="
         + "&start=1&results=" + aNumResults
         + "&region=" + region
         + "&fr=flock-search";
}

YahooLiveSearch.prototype._getResults =
function YLS__getResults(aXMLDoc, aMaxResults) {
  var unescapeHTML = CC["@mozilla.org/feed-unescapehtml;1"]
                     .getService(CI.nsIScriptableUnescapeHTML);

  var items = aXMLDoc.getElementsByTagName("Result");

  var results = [];
  for (var i = 0; i < items.length; i++) {
    var item = items[i];

    var title = item.getElementsByTagName("Title")[0].firstChild.nodeValue;
    var url = item.getElementsByTagName("Url")[0].firstChild.nodeValue;

    results.push({ title: unescapeHTML.unescape(title), url: url });
  }

  return results;
}

var NSGetModule =
  FlockLiveSearchProvider.generateNSGetModule(YAHOO_LIVE_SEARCH_NAME,
                                              YahooLiveSearch);
