// vim: ts=2 sw=2 expandtab
//
// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

// Developer ID = 1aiV9CsTs3o
// Developer Secret = flock_is_open_source

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

CU.import("resource:///modules/FlockScheduler.jsm");
CU.import("resource:///modules/FlockSvcUtils.jsm");
CU.import("resource:///modules/FlockXPCOMUtils.jsm");
FlockXPCOMUtils.debug = false;
CU.import("resource://gre/modules/JSON.jsm");

const YT_DEVID = "1aiV9CsTs3o";
const YT_DEVSECRET = "flock_is_open_source";

const MODULE_NAME = "Youtube";
const CLASS_ID = Components.ID("{DCB6A01E-7D4A-4C30-AB3D-9CC98C02F617}");
const CONTRACT_ID = "@flock.com/?photo-api-youtube;1";
const CLASS_NAME = "YouTube";
const CLASS_SHORT_NAME = "youtube";

const YOUTUBE_FAVICON = "chrome://flock/content/services/youtube/favicon.png";
const SERVICE_ENABLED_PREF = "flock.service.youtube.enabled";
const FRIENDS_AVATAR_TIMEOUT = 5000;  // in milliseconds
const FRIENDS_ACTIVITY_TIMEOUT = 500;  // in milliseconds

const FAV_CHANNELS_ROOT = "urn:media:favorites";

const FLOCK_RDNDS_CONTRACTID = "@flock.com/rich-dnd-service;1";

const YOUTUBE_IDENTITY_URN_PREFIX = "urn:flock:identity:youtube:";
const PEOPLE_PROPERTIES = "chrome://flock/locale/people/people.properties";
const YOUTUBE_STRING_BUNDLE = "chrome://flock/locale/services/youtube.properties";

// From nsIXMLHttpRequest.idl
// 4: COMPLETED     Finished with all operations.
const XMLHTTPREQUEST_READYSTATE_COMPLETED = 4;

const HTTP_CODE_OK = 200;

// The delay between two refreshes when the sidebar is closed (in seconds)
const YOUTUBE_REFRESH_INTERVAL = 3600; // 60 minutes


function _getIdentityUrn(aAccountId, aUid) {
  var result = YOUTUBE_IDENTITY_URN_PREFIX
             + aAccountId + ":"
             + aUid;
  return result;
}

function loadLibraryFromSpec(aSpec) {
  var loader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                         .getService(Components.interfaces.mozIJSSubScriptLoader);
  loader.loadSubScript(aSpec);
}

// String defaults... may be updated later through Web Detective
var gStrings = {
  "domains": "youtube.com",
  "userlogin": "http://www.youtube.com/login",
  "userprofile": "http://www.youtube.com/profile?user=%accountid%",
  "editprofile": "http://youtube.com/my_profile",
  "inbox": "http://youtube.com/my_messages",
  "noavatarimage": "no_videos_140.jpg"
};

var flockMediaItemFormatter = {
  canBuildTooltip: true,
  canBuildHTML: true,
  canBuildLargeHTML: true,
  canBuildBBCode: true,
  canBuildMiniPage: true,

  buildTooltip: function fmif_buildTooltip(aMediaItem) {

    default xml namespace =
      "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

    var extras = <vbox anonid="ttUserTxt" />;
    if (aMediaItem.length_seconds) {
      extras.appendChild(<label>Length: {aMediaItem.length_seconds} seconds</label>);
    }
    if (aMediaItem.rating_avg) {
      extras.appendChild(<label>Rating: {aMediaItem.rating_avg}
                         ({aMediaItem.rating_count} times)</label>);
    }
    if (aMediaItem.view_count) {
      extras.appendChild(<label>Views: {aMediaItem.view_count}</label>);
    }

    extras.appendChild(<label class="user">{aMediaItem.username}</label>);

    var xml =
      <vbox>
        <hbox>
          <image src={aMediaItem.midSizePhoto} style="margin-bottom: 2px;" />
          <spacer flex="1" />
        </hbox>
        <vbox anonid="ttInfoContainer">
          <label anonid="ttTitleTxt">{aMediaItem.title}</label>
          {extras}
        </vbox>
      </vbox>;

    return xml;
  },

  buildHTML: function fmif_buildHTML(aMediaItem) {
    var flashUrl = aMediaItem.webPageUrl.replace(/\/\?v\=/,"/v/");
    var html = '<object width="425" height="350">'
         + '<param name="movie" value="'
         + flashUrl
         + '"/>'
         + '<embed src="'
         + flashUrl
         + '" type="application/x-shockwave-flash" width="425" height="350"/>'
         + aMediaItem.webPageUrl
         + '</object>';
    return html;
  },

  buildLargeHTML: function fmif_buildLargeHTML(aMediaItem) {
    return this.buildHTML(aMediaItem);
  },

  buildBBCode: function fmif_buildBBCode(aMediaItem) {
    var flashUrl = aMediaItem.webPageUrl.replace(/\/\?v\=/,"/watch?v=");
    var bbCode = "[youtube]" + flashUrl + "[/youtube]";
    return bbCode;
  },

  buildMiniPage: function fmif_buildMiniPage(aMediaItem) {
    var flashUrl = aMediaItem.webPageUrl.replace(/\/\?v\=/,"/v/");
    var html = '<html><head><title>'
             + aMediaItem.title
             + ' (' + aMediaItem.username + ')</title></head>'
             + '<body><center><object width="425" height="350">'
             + '<param name="movie" value="'
             + flashUrl
             + '"/>'
             + '<embed src="'
             + flashUrl
             + '&autoplay=1" type="application/x-shockwave-flash"'
             + ' width="425" height="350"/></object></center></body></html>';
    return html;
  }
};


/*************************************************************************
 * Component: YoutubeService
 *************************************************************************/

function YoutubeService() {
  loadLibraryFromSpec("chrome://flock/content/photo/photoAPI.js");
  FlockSvcUtils.getACUtils(this);
  FlockSvcUtils.getLogger(this);

  FlockSvcUtils.flockIWebService.addDefaultMethod(this, "isHidden");
  FlockSvcUtils.flockIWebService.addDefaultMethod(this, "getStringBundle");
  this._loginURL = gStrings["userlogin"];
  FlockSvcUtils.flockIWebService.addDefaultMethod(this, "loginURL");

  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "getAccount");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "getAuthenticatedAccount");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "getAccounts");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "removeAccount");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "docRepresentsSuccessfulLogin");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "getAccountIDFromDocument");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "getCredentialsFromForm");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "ownsDocument");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "ownsLoginForm");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "getSessionValue");
  FlockSvcUtils.flockILoginWebService
               .addDefaultMethod(this, "logout");

  FlockSvcUtils.flockIMediaWebService
               .addDefaultMethod(this, "getChannel");
  FlockSvcUtils.flockIMediaWebService
               .addDefaultMethod(this, "enumerateChannels");
  FlockSvcUtils.flockIMediaWebService
               .addDefaultMethod(this, "getIconForQuery");

  FlockSvcUtils.flockIRichContentDropHandler
               .addDefaultMethod(this, "_handleTextareaDrop");

  FlockSvcUtils.flockIMediaEmbedWebService
               .addDefaultMethod(this, "getSharingContent");

  this._obs = CC["@mozilla.org/observer-service;1"]
              .getService(CI.nsIObserverService);
  this._obs.addObserver(this, "xpcom-shutdown", false);

  this._ppSvc = CC["@flock.com/people-service;1"]
                .getService(CI.flockIPeopleService);
  this.url = "http://www.youtube.com";
  this._currentQueryExhausted = false;

  var sbs = CC["@mozilla.org/intl/stringbundle;1"]
            .getService(CI.nsIStringBundleService);
  var bundle = sbs.createBundle(YOUTUBE_STRING_BUNDLE);

  this._channels = {
    "special:added": {
      title: bundle.GetStringFromName("flock.youtube.title.added"),
      supportsSearch: false,
      feed: "http://youtube.com/rss/global/recently_added.rss"
    },
    "special:featured": {
      title: bundle.GetStringFromName("flock.youtube.title.featured"),
      supportsSearch: false
    },
    "special:top_favorites": {
      title: bundle.GetStringFromName("flock.youtube.title.topfav"),
      supportsSearch: false,
      feed: "http://youtube.com/rss/global/top_favorites.rss"
    },
    "special:rated": {
      title: bundle.GetStringFromName("flock.youtube.title.rated"),
      supportsSearch: false,
      feed: "http://youtube.com/rss/global/top_rated.rss"
    }
  };

  var prefs = CC["@mozilla.org/preferences-service;1"]
              .getService(CI.nsIPrefBranch);
  if (prefs.getPrefType(SERVICE_ENABLED_PREF) &&
     !prefs.getBoolPref(SERVICE_ENABLED_PREF))
  {
    this._logger.debug(SERVICE_ENABLED_PREF + " is FALSE! Not initializing.");
    this.deleteCategories();
    return;
  }

  this._accountClassCtor = YoutubeAccount;

  FlockSvcUtils.getCoop(this);

  this._c_svc = new this._coop.Service(
    "urn:youtube:service",
    {
      name: "youtube",
      desc: "The YouTube Service"
    }
  );
  this._c_svc.serviceId = CONTRACT_ID;

  // Load Web Detective file
  this._WebDetective = this._acUtils.useWebDetective("youtube.xml");
  for (var s in gStrings) {
    gStrings[s] = this._WebDetective.getString("youtube", s, gStrings[s]);
  }
  this._c_svc.domains = gStrings["domains"];

  this.urn = this._c_svc.id();

  this._updateFriendAvatarsComplete = false;
  
  this._updateFriendActivityComplete = false;
  this._friendActivityArray = [];

  // Initialize member that specifies path for localized string bundle
  this._stringBundlePath = YOUTUBE_STRING_BUNDLE;
}

YoutubeService.prototype = new FlockXPCOMUtils.genericComponent(
  CLASS_NAME,
  CLASS_ID,
  CONTRACT_ID,
  YoutubeService,
  CI.nsIClassInfo.SINGLETON,
  [
    CI.flockILoginWebService,
    CI.flockIMediaEmbedWebService,
    CI.flockIMediaWebService,
    CI.flockIPollingService,
    CI.flockIRichContentDropHandler,
    CI.flockISocialWebService,
    CI.flockIWebService,
    CI.flockIYoutubeService,
    CI.nsIObserver
  ]
);

YoutubeService.prototype._xpcom_categories = [
  { category: "wsm-startup" },
  { category: "flockWebService", entry: CLASS_SHORT_NAME },
  { category: "flockMediaProvider", entry: CLASS_SHORT_NAME },
  { category: "flockRichContentHandler", entry: CLASS_SHORT_NAME }  
];

/**************************************************************************
 * YoutubeService: flockILoginWebService Implementation
 **************************************************************************/

// readonly attribute AString domains;
YoutubeService.prototype.__defineGetter__("domains",
function YoutubeService_getdomains() {
  this._logger.debug("Getting property: domains");
  return this._c_svc.domains;
});

// readonly attribute boolean needPassword;
YoutubeService.prototype.needPassword = false;

/**
 * flockIWebServiceAccount addAccount(in AString aAccountID,
 *                                    in boolean aIsTransient,
 *                                    in flockIListener aFlockListener);
 * @see flockILoginWebService#addAccount
 */
YoutubeService.prototype.addAccount =
function YoutubeService_addAccount(aAccountID, aIsTransient, aFlockListener) {
  this._logger.debug(".addAccount('" + aAccountID + "', "
                                     + aIsTransient + ", ...)");

  var url = gStrings["userprofile"].replace("%accountid%", aAccountID);
  var accountURN = this._acUtils.createAccount(this,
                                               aAccountID,
                                               aAccountID,
                                               url,
                                               aIsTransient);
  var c_account = this._coop.get(accountURN);

  if (!c_account.friendsList) {
    var friendsListUrn = accountURN + ":friends";
    var friendsList = new this._coop.FriendsList(
      friendsListUrn,
      {
        account: c_account
      }
    );
    c_account.friendsList = friendsList;
  }

  // Instantiate account component
  var account = this.getAccount(accountURN);
  account._addMediaToFavorites(aAccountID);
  if (aFlockListener) {
    aFlockListener.onSuccess(account, "addAccount");
  }
  return account;
};

// DEFAULT: boolean docRepresentsSuccessfulLogin(in nsIDOMHTMLDocument aDocument);
// DEFAULT: AString getAccountIDFromDocument(in nsIDOMHTMLDocument aDocument);
// DEFAULT: nsISimpleEnumerator getAccounts();
// DEFAULT: flockIWebServiceAccount getAccount(in AString aUrn);
// DEFAULT: flockIWebServiceAccount getAuthenticatedAccount();
// DEFAULT: nsILoginInfo getCredentialsFromForm(in nsIDOMHTMLFormElement aForm);
// DEFAULT: AString getSessionValue();

// DEFAULT: boolean ownsDocument(in nsIDOMHTMLDocument aDocument);
// DEFAULT: boolean ownsLoginForm(in nsIDOMHTMLFormElement aForm);
// DEFAULT: void removeAccount(in AString aUrn);

/**
 * void updateAccountStatusFromDocument(in nsIDOMHTMLDocument aDocument,
 *                                      in AString aAcctURN,
 *                                      in flockIListener aFlockListener);
 * @see flockILoginWebService#updateAccountStatusFromDocument
 */
YoutubeService.prototype.updateAccountStatusFromDocument =
function YoutubeService_updateAccountStatusFromDocument(aDocument,
                                                        aAcctURN,
                                                        aFlockListener)
{
  this._logger.info(".updateAccountStatusFromDocument('" + aDocument.URL + "', '"
                                                         + aAcctURN + "', ...)");
  if (aAcctURN) {
    // We're logged in to this account
    var account = this.getAccount(aAcctURN);
    if (!account.isAuthenticated()) {
      var inst = this; 
      var flockListener = {
        onSuccess: function flockListener_onSuccess(aSubject, aTopic) {
          inst._logger.debug(".updateAccountStatusFromDocument(): "
                             + "loginListener: onSuccess()");
          account.coopObj.isAuthenticated = true;
          inst._updateFriendAvatarsComplete = false;
          if (aFlockListener) {
            aFlockListener.onSuccess(account, null);
          }
        },
        onError: function flockListener_onError(aFlockError, aTopic) {
          inst._logger.debug(".updateAccountStatusFromDocument(): "
                             + "loginListener: onError()");
          inst._logger.debug(aFlockError ? (aFlockError.errorString)
                                         : "No details");
        }
      };
      account.login(flockListener);
    }
  } else if (this._WebDetective
                 .detect("youtube", "loggedout", aDocument, null))
  {
    // We're logged out (of all accounts)
    this._acUtils.markAllAccountsAsLoggedOut(CONTRACT_ID);
  }
};


/**************************************************************************
 * YoutubeService: flockIMediaEmbedWebService Implementation
 **************************************************************************/

/**
 * boolean checkIsStreamUrl(in AString aUrl);
 * @see flockIMediaEmbedWebService#checkIsStreamUrl
 */
YoutubeService.prototype.checkIsStreamUrl =
function YoutubeService_checkIsStreamUrl(aUrl) {
  if (this._WebDetective.detectNoDOM("youtube", "isStreamUrl", "", aUrl, null)) {
    this._logger.debug("Checking if url is youtube stream: YES: " + aUrl);
    return true;
  }
  this._logger.debug("Checking if url is youtube stream: NO: " + aUrl);
  return false;
};

/**
 * void getMediaQueryFromURL(in AString aUrl, in flockIListener aFlockListener);
 * @see flockIMediaEmbedWebService#getMediaQueryFromURL
 */
YoutubeService.prototype.getMediaQueryFromURL =
function youTubeService_getMediaQueryFromURL(aUrl, aFlockListener) {
  var videoId = this.getVideoIDFromUrl(aUrl);
  if (videoId) {
    var inst = this;
    var myListener = {
      onResult: function getMediaQueryFromURL_onResult(aXML) {
        var userID = aXML.getElementsByTagName('author')[0].firstChild.nodeValue;
        var results = CC["@mozilla.org/hash-property-bag;1"]
                      .createInstance(CI.nsIWritablePropertyBag2);
        results.setPropertyAsAString("query", "user:" + userID + "|username:" + userID);
        // Note: The "title" property is currently not used by the media bar.
        results.setPropertyAsAString("title", userID);
        aFlockListener.onSuccess(results, "query");
      },
      onError: function getMediaQueryFromURL_onError(aFlockError) {
        inst._logger.error(".getMediaQueryFromURL() myListener.onError()");
        aFlockListener.onError(aFlockError, null);
      }
    };
    var params = {};
    params.video_id = videoId;
    this.call(myListener, "youtube.videos.get_details", params);
  } else {
    this._logger.error(".getMediaQueryFromURL() Unable to get user.");
    aFlockListener.onError(null, "Unable to get user.");
  }
};

// DEFAULT: boolean getSharingContent(in nsIDOMHTMLElement aSrc,
//                                    in nsIWritablePropertyBag2 aProp);


/**************************************************************************
 * YoutubeService: flockIMediaWebService Implementation
 **************************************************************************/

// readonly attribute boolean supportsUsers;
YoutubeService.prototype.supportsUsers = true;

// DEFAULT: flockIMediaChannel getChannel(in AString aChannelId);

/**
 * void decorateForMedia(in nsIDOMHTMLDocument aDocument);
 * @see flockIMediaWebService#decorateForMedia
 */
YoutubeService.prototype.decorateForMedia =
function YoutubeService_decorateForMedia(aDocument) {
  aDocument.QueryInterface(Components.interfaces.nsIDOMHTMLDocument);
  this._logger.debug(".decorateForMedia('" + aDocument.URL + "')");
  var results = FlockSvcUtils.newResults();
  if (this._WebDetective.detect("youtube", "media", aDocument, results)) {
    var sbs = CC["@mozilla.org/intl/stringbundle;1"]
              .getService(CI.nsIStringBundleService);
    var bundle = sbs.createBundle(YOUTUBE_STRING_BUNDLE);

    var mediaArr = [];
    var userid = results.getPropertyAsAString("userid");

    // media item for user videos
    var vidname = bundle.formatStringFromName("flock.youtube.videosUC",
                                              [userid], 1);
    var media = {
      name: userid,
      query: "user:" + userid + "|username:" + userid,
      label: vidname,
      favicon: this.icon,
      service: this.shortName
    };
    mediaArr.push(media);

    // media item for user favorites
    var favname = bundle.formatStringFromName("flock.youtube.favoritesUC",
                                              [userid], 1);
    media = {
      name: userid,
      query: "favorites:" + userid,
      label: favname,
      favicon: this.icon,
      service: this.shortName
    };
    mediaArr.push(media);

    if (!aDocument._flock_decorations) {
      aDocument._flock_decorations = {};
    }

    if (aDocument._flock_decorations.mediaArr) {
      aDocument._flock_decorations.mediaArr =
        aDocument._flock_decorations.mediaArr.concat(mediaArr);
    } else {
      aDocument._flock_decorations.mediaArr = mediaArr;
    }

    this._obs.notifyObservers(aDocument, "media", "media:update");
  }
};

// DEFAULT: nsISimpleEnumerator enumerateChannels();

/**
 * void findByUsername(in flockIListener aFlockListener,
 *                     in AString aUsername);
 * @see flockIMediaWebService#findByUsername
 */
YoutubeService.prototype.findByUsername =
function YoutubeService_findByUsername(aFlockListener, aUsername) {
  var inst = this;
  var myListener = {
    onResult: function findByUsername_onResult(aXML) {
      inst._logger.debug("Auth XML is:\n" + aXML + "\n");
      // XXX ERWAN: why are we making a call to the server, if we don't
      // use the xml we get as result??
      // XXX marcus: Debugging shows that aXML is empty.
      var newUserObj = CC["@flock.com/person;1"]
                       .createInstance(CI.flockIPerson);
      newUserObj.serviceId = inst.contractID;
      newUserObj.accountId = aUsername;
      newUserObj.name = aUsername;
      newUserObj.totalMedia = -1; // See mediabar.js
      aFlockListener.onSuccess(newUserObj, "");
    },
    onError: function findByUsername_onError(aXML) {
      inst._logger.error(".findByUsername() myListener.onError()");
      aFlockListener.onError(null, null);
    }
  };
  var params = {};
  params.user = aUsername;
  this.call(myListener, "youtube.users.get_profile", params);
};

/**
 * void getAlbums(in flockIListener aFlockListener,
 *                in AString aUsername);
 * @see flockIMediaWebService#getAlbums
 */
YoutubeService.prototype.getAlbums =
function YoutubeService_getAlbums(aFlockListener, aUsername) { 
  aFlockListener.onError(null, null);
};

/**
 * flockIMediaItemFormatter getMediaItemFormatter();
 * @see flockIMediaWebService#getMediaItemFormatter
 */
YoutubeService.prototype.getMediaItemFormatter =
function YoutubeService_getMediaItemFormatter() {
  return flockMediaItemFormatter;
};

/**
 * void migrateAccount(in AString aId, in AString aUsername);
 * @see flockIMediaWebService#migrateAccount
 */
YoutubeService.prototype.migrateAccount =
function YoutubeService_migrateAccount(aId, aUsername) {
  this.addAccount(aId, false, null);
};

/**
 * void search(in flockIListener aFlockListener,
 *             in AString aQuery,
 *             in long aCount,
 *             in long aPage,
 *             in AString aRequestId);
 * @see flockIMediaWebService#search
 */
YoutubeService.prototype.search =
function YoutubeService_search(aFlockListener,
                               aQueryString,
                               aCount,
                               aPage,
                               aRequestId)
{
  if (aPage == 1) {
    // Same search refreshed or new search
    this._currentQueryExhausted = false;
  }
  if (this._currentQueryExhausted) {
    // No need to search as all pages for this query have been returned
    return;
  }

  var aQuery = new queryHelper(aQueryString);
  if (aQuery.favorites && !aQuery.user) {
    aQuery.user = aQuery.favorites;
  }
  if (!aQuery.user && aQuery.special != "favorites") {
    this._queryChannel(aFlockListener, aQueryString, aCount, aPage, aRequestId);
    return;
  }

  var aUserid = aQuery.user;
  var params = {
    user: aUserid
  };

  var inst = this;
  var myListener = {
    onResult: function search_onResult(aXML) {
      inst._logger.debug(".search() myListener.onResult()");
      var rval = inst._handlePhotosResult(aXML, aUserid);

      if (rval.length < aCount) {
        // All results for the current query have been returned 
        inst._currentQueryExhausted = true;
      }

      var resultsEnum = {
        hasMoreElements: function resultsEnum_hasMoreElements() {
          return (rval.length > 0);
        },
        getNext: function resultsEnum_getNext() {
          return rval.shift();
        },
        QueryInterface: function resultsEnum_QueryInterface(aIID) {
          if (aIID.equals(CI.nsISimpleEnumerator)) {
            return this;
          }
          throw CR.NS_ERROR_NO_INTERFACE;
        }
      };
      aFlockListener.onSuccess(resultsEnum, aRequestId);
    },
    onError: function search_onError(aFlockError) {
      inst._logger.error(".search() myListener.onError()");
      aFlockListener.onError(aFlockError, aRequestId);
    }
  };

  if (aQuery.favorites) {
    this.call(myListener, "youtube.users.list_favorite_videos", params);
  } else {
    params.page = aPage;
    params.per_page = aCount;
    this.call(myListener, "youtube.videos.list_by_user", params);
  }
};

/**
 * boolean supportsSearch(in AString aQuery);
 * @see flockIMediaWebService#supportsSearch
 */
YoutubeService.prototype.supportsSearch =
function YoutubeService_supportsSearch(aQueryString) {
  var aQuery = new queryHelper(aQueryString);

  if (aQuery.special) {
    var channel = this._channels["special:" + aQuery.special];
    if (channel) {
      return channel.supportsSearch;
    }
  }

  // None of the other apis/feeds support search
  return false;
};

YoutubeService.prototype._getFeed =
function YoutubeService__getFeed(aFlockListener, aFeedURL, aRequestId) {
  var ios = CC["@mozilla.org/network/io-service;1"].getService(CI.nsIIOService);
  var uri = ios.newURI(aFeedURL, null, null);

  var inst = this;
  var feedListener = {
    onGetFeedComplete: function oncomplete(feed) {
      inst._handleFeedResult(aFlockListener, feed, aRequestId);
    },
    onError: function onError(aFlockError) {
      aFlockListener.onError(aFlockError, aRequestId);
    }
  };

  var fm = Components.classes["@flock.com/feed-manager;1"]
    .getService(Components.interfaces.flockIFeedManager);
  fm.getFeedBypassCache(uri, feedListener, false);
};

YoutubeService.prototype._handleFeedResult =
function YoutubeService__handleFeedResult(aFlockListener, aFeed, aRequestId) {
  var photos = [];
  var nextIdx = 0;

  var items = aFeed.getItems();
  while (items && items.hasMoreElements()) {
    var item = items.getNext();

    var newMediaItem = CC["@flock.com/photo;1"]
                       .createInstance(CI.flockIMediaItem);
    newMediaItem.init(this.shortName, flockMediaItemFormatter);
    newMediaItem.has_miniView = "true";

    newMediaItem.title = item.getTitle();
    newMediaItem.webPageUrl = item.getLink().spec;

    newMediaItem.uploadDate = item.getPubDate();
    // Undo the scaling applied in _handlePhotosResult() and ensure that the ID
    // is truly unique so tooltips work properly.
    newMediaItem.id = "" + (newMediaItem.uploadDate / 1000) + nextIdx++;

    var author = item.getAuthor();
    newMediaItem.username = author;
    newMediaItem.userid = author;

    var desc = item.getContent();
    var re = /<img [^>]*src="([^"]*)"/i;
    var thumbnail = desc.match(re)[1];
    newMediaItem.thumbnail = thumbnail;
    newMediaItem.midSizePhoto = thumbnail;
    newMediaItem.largeSizePhoto = thumbnail;

    newMediaItem.is_public = "true";
    newMediaItem.is_video = true;

    photos.push(newMediaItem);
  }

  var resultsEnum = {
    hasMoreElements: function resultsEnum_hasMoreElements() {
      return (photos.length > 0);
    },
    getNext: function resultsEnum_getNext() {
      return photos.shift();
    },
    QueryInterface: function resultsEnum_QueryInterface(aIID) {
      if (aIID.equals(CI.nsISimpleEnumerator)) {
        return this;
      }
      throw CR.NS_ERROR_NO_INTERFACE;
    }
  };
  aFlockListener.onSuccess(resultsEnum, aRequestId);
};

YoutubeService.prototype._handlePhotosResult =
function YoutubeService__handlePhotoResult(aXML, aUserid) {
  var rval = [];
  var nextIdx = 0;
  var photoList = aXML.getElementsByTagName("video");
  for (var i = 0; i < photoList.length; i++) {
    var photo = photoList[i];
    var newMediaItem = CC["@flock.com/photo;1"]
                       .createInstance(CI.flockIMediaItem);
    newMediaItem.init(this.shortName, flockMediaItemFormatter);
    newMediaItem.has_miniView = "true";
    for (var j = 0; j < photo.childNodes.length; j++) {
      var child = photo.childNodes[j];
      var contentNode = child.childNodes[0];
      var childContent = "";
      if (contentNode) {
        childContent = contentNode.nodeValue;
      }
      switch (child.tagName) {
        case "author":
          newMediaItem.username = childContent;
          newMediaItem.userid = childContent;
          break;
        // case "id":
          // newMediaItem.id = childContent;
          // break;
        case "title":
          newMediaItem.title = childContent;
          break;
        case "upload_time":
          newMediaItem.uploadDate = childContent * 1000;
          newMediaItem.id = "" + parseInt(childContent) + nextIdx++;
          break;
        case "length_seconds":
          newMediaItem.length_seconds = childContent;
          break;
        case "rating_avg":
          newMediaItem.rating_avg = childContent;
          break;
        case "description":
          newMediaItem.description = childContent;
          break;
        case "comment_count":
          newMediaItem.comment_count = childContent;
          break;
        case "view_count":
          newMediaItem.view_count = childContent;
          break;
        case "rating_count":
          newMediaItem.rating_count = childContent;
          break;
        case "url":
          newMediaItem.webPageUrl = childContent;
          break;
        case "thumbnail_url":
          newMediaItem.thumbnail = childContent;
          newMediaItem.midSizePhoto = childContent;
          newMediaItem.largeSizePhoto = childContent;
          break;
      }
    }

    newMediaItem.is_public = "true";
    newMediaItem.is_video = true;
    rval.push(newMediaItem);
  }
  return rval;
};

YoutubeService.prototype._queryChannel =
function YoutubeService__queryChannel(aFlockListener,
                                      aQueryString,
                                      aCount,
                                      aPage,
                                      aRequestId)
{
  var aQuery = new queryHelper(aQueryString);
  // return; // XXX TODO FIXME: This is killing perf

  var inst = this;
  var myListener = {
    onResult: function _queryChannel_onResult(aXML) {
      var rval = inst._handlePhotosResult(aXML);

      if (rval.length < aCount) {
        // All results for the current query have been returned
        inst._currentQueryExhausted = true;
      }

      var resultsEnum = {
        hasMoreElements: function resultsEnum_hasMoreElements() {
          return (rval.length > 0);
        },
        getNext: function resultsEnum_getNext() {
          return rval.shift();
        },
        QueryInterface: function resultsEnum_QueryInterface(aIID) {
          if (aIID.equals(CI.nsISimpleEnumerator)) {
            return this;
          }
          throw CR.NS_ERROR_NO_INTERFACE;
        }
      };
      aFlockListener.onSuccess(resultsEnum, aRequestId);
    },
    onError: function _queryChannel_onError(aFlockError) {
      inst._logger.error("._queryChannel() myListener.onError()");
      aFlockListener.onError(aFlockError, aRequestId);
    }
  };

  var params = {};
  if (aQuery.search) {
    params.tag = aQuery.search;
    params.page = aPage;
    params.per_page = aCount;
    this.call(myListener, 'youtube.videos.list_by_tag', params);
  } else {
    // This API call doesn't support pagination and only shows most recent 25 items
    if (aPage > 1) {
      return;
    }
    var channel = this._channels[aQuery.stringVal];
    if (!channel) {
      return;
    }
    if (channel.feed) {
      this._getFeed(aFlockListener, channel.feed, aRequestId);
    } else {
      this.call(myListener, "youtube.videos.list_featured", params);
    }
  }
};


/**************************************************************************
 * YoutubeService: flockIPollingService Implementation
 **************************************************************************/

/**
 * void refresh(in AString aURN, in flockIPollingListener aPollingListener);
 * @see flockIPollingService#refresh
 */
YoutubeService.prototype.refresh =
function YoutubeService_refresh(aURN, aPollingListener) {
  this._logger.info(".refresh('" + aURN + "', ...)");
  var refreshItem = this._coop.get(aURN);

  if (refreshItem.isInstanceOf(this._coop.Account)) {
    this._logger.debug("refreshing an Account");
    if (refreshItem.isAuthenticated) {
      this._refreshAccount(refreshItem, aPollingListener);
    } else {
      // If the user is not logged in, return a success without
      // refreshing anything
      aPollingListener.onResult();
    }
  } else {
    throw CR.NS_ERROR_ABORT;
  }
};

YoutubeService.prototype._addCoopPerson =
function YoutubeService__addCoopPerson(aPhotoPerson, aCoopAccount) {
  var person = aPhotoPerson;
  var personURL = this._WebDetective.getString("youtube", "friendprofile", "")
                                   .replace("%accountid%", person.accountId);
  var identityUrn = _getIdentityUrn(aCoopAccount.accountId,
                                    person.accountId);
  var identity = this._coop.get(identityUrn);
  if (identity) {
    if (this._personUpdateRequired(aCoopAccount,person)) {
      // Update data of the identity coop obj here
      identity.name = person.accountId;
      identity.URL = personURL;
      this._incrementMedia(identity, person);
    }
  } else {
    identity = new this._coop.Identity(
      identityUrn,
      {
        name: person.accountId,
        serviceId: CONTRACT_ID,
        accountId: person.accountId,
        avatar: "",
        statusMessage: "",
        lastUpdateType: "media",
        lastUploadedMedia: person.videoCount,
        URL: personURL
      }
    );
    aCoopAccount.friendsList.children.add(identity);
    identity.lastUpdate = 0; // triggers the RDF observers
    this._addIdentityToActivityArray(identity);
  }
};

YoutubeService.prototype._addIdentityToActivityArray =
function YoutubeService__addIdentityToActivityArray(aIdentity) {
  this._logger.info("._addIdentityToActivityArray('"+ aIdentity.name + "')");
  
   //only want to add to array when friendsActivity timer is ready
  if (!this._updateFriendActivityComplete) {
    this._friendActivityArray.push(aIdentity);
  }
};

/**
 * The avatar URL we scraped may be a relative URL. If so, prepend
 * "http://www.youtube.com/" to it.
 */
YoutubeService.prototype._fixRelativeAvatarUrl = 
function YoutubeService__fixRelativeAvatarUrl(aUrl) {
  var url = aUrl;

  if (url.indexOf("http") != 0) {
    url = "http://www.youtube.com";
    // Do we need to add a leading slash?
    if (aUrl.indexOf("/") != 0) {
      url += "/";
    }
    url += aUrl;
  }

  return url;
};

YoutubeService.prototype._handleFriendsResult =
function YoutubeService__handleFriendsResult(aXML,
                                             aCoopAcct,
                                             aRefreshListener)
{
  this._logger.debug("._handleFriendsResult(...)");
  
  var friendList = aXML.getElementsByTagName("friend");
  var inst = this;
  var peopleHash = [];
  this._logger.debug(" - found " + friendList.length + " friends");
  for (var i = 0; i < friendList.length; i++) {
    var friend = friendList[i];
    var friendObj = {};
    for (var j = 0; j < friend.childNodes.length; j++) {
      var child = friend.childNodes[j];
      var contentNode = child.childNodes[0];
      var childContent = "";
      if (contentNode) {
        childContent = contentNode.nodeValue;
      }
      switch (child.tagName) {
        case "user":
          friendObj.accountId = childContent;
          friendObj.name = childContent;
          break;
        case "video_upload_count":
          friendObj.videoCount = childContent;
          break;
      }
    }
    peopleHash[friendObj.accountId] = friendObj;
    peopleHash[friendObj.accountId].media = {
      count: 1,
      latest: friendObj.videoCount
    };
  }
  function myWorker(aShouldYield) {
    if (aCoopAcct.isAuthenticated) {
      // REMOVE locally people removed on the server
      var localEnum = aCoopAcct.friendsList.children.enumerate();
      while (localEnum.hasMoreElements()) {
        var identity = localEnum.getNext();
        if (!peopleHash[identity.accountId]) {
          inst._logger.info("Friend " + identity.accountId
                                      + " has been deleted on the server");
          aCoopAcct.friendsList.children.remove(identity);
          identity.destroy();
        }
      }
      // ADD or update existing people
      for (var uid in peopleHash) {
        if (!aCoopAcct.isAuthenticated) {
          // If this account is no longer authenticated, bail out.
          return;
        }
        inst._addCoopPerson(peopleHash[uid], aCoopAcct);
        if (aShouldYield()) {
          yield;
          if (!aCoopAcct.isAuthenticated) {
            // Account has just been deleted or logged out
            break;
          }
        }
      }
    }
    
    inst._updateFriendRecentActivity(aCoopAcct, aRefreshListener);
    inst._updateFriendAvatars(aCoopAcct, aRefreshListener);
  } // end myWorker()
    
  FlockScheduler.schedule(null, 0.05, 10, myWorker);
};

YoutubeService.prototype._incrementMedia =
function YoutubeService__incrementMedia(identity, person) {
  this._logger.debug("._incrementMedia('" + identity.id() + ","
                                          + person.accountId + "')");
  
  if (identity.lastUploadedMedia != person.videoCount) {
    identity.lastUploadedMedia = person.videoCount;
    identity.unseenMedia = 1;
    this._addIdentityToActivityArray(identity);
  }
};

YoutubeService.prototype._personUpdateRequired =
function YoutubeService__personUpdateRequired(aCoopPerson, aPerson) {
  return (aCoopPerson.name != aPerson.accountId);
};

YoutubeService.prototype._refreshAccount =
function YoutubeService__refreshAccount(aCoopAcct, aPollingListener) {
  var inst = this;
  var pplRefreshListener = CC["@flock.com/people-refresh-listener;1"]
                           .createInstance(CI.flockIPeopleRefreshListener);
  pplRefreshListener.init(3, aCoopAcct.id(), aPollingListener, 0);

  // Get new message count
  var hr = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
           .createInstance(CI.nsIXMLHttpRequest);
  hr.onreadystatechange = function updateUserAccount_onreadystatechange(aEvent)
  {
    if (hr.readyState == XMLHTTPREQUEST_READYSTATE_COMPLETED) {
      var results = FlockSvcUtils.newResults();
      if (inst._WebDetective.detectNoDOM("youtube", "accountinfo", "",
                                         hr.responseText, results))
      {
        // Set the message count
        var messages = "0";
        try {
          messages = results.getPropertyAsAString("messages");
          if (!messages || !messages.length) {
            messages = "0";
          }
        } catch (ex) {}
        
        inst._logger.debug("aCoopAcct.accountMessages " + messages);
        if (aCoopAcct.accountMessages < messages) {
          inst._ppSvc.togglePeopleIcon(true);
        }
        aCoopAcct.accountMessages = messages;

        // Set the avatar for the mecard.
        // If avatar returned as YT no avatar, then set coop avatar to
        // null and let people sidebar code set the Flock no avatar img.
        var fixedAvatar = null;
        try {
          var avatar = results.getPropertyAsAString("avatar");
          if (avatar.match(gStrings["noavatarimage"])) {
            avatar = null;
            inst._logger.debug("no avatar for account, setting to null\n");
          } else {
            fixedAvatar = inst._fixRelativeAvatarUrl(avatar);
            inst._logger.debug("aCoopAcct.accountAvatar " + aCoopAcct.avatar);
          }
        } catch (ex) {}
        aCoopAcct.avatar = fixedAvatar;
      } else {
        aCoopAcct.accountMessages = 0;
      }
      pplRefreshListener.onSuccess(null, "updateUserAccount");
    }
  };

  var up = gStrings["userprofile"];
  hr.open("GET", up.replace("%accountid%", aCoopAcct.accountId));
  // get friends
  var getFriendsListener = {
    onResult: function getFriendsListener_onResult(aXML) {
      if (aCoopAcct.isAuthenticated) {
        inst._handleFriendsResult(aXML, aCoopAcct, pplRefreshListener);
        hr.send(null);
      } else {
        aPollingListener.onResult();
      }
    },
    onError: function getFriendsListener_onError(aXML) {
      inst._logger.error("Error on getFriendsListener");
      aPollingListener.onError(null);
    }
  };
  var params = {};
  params.user = aCoopAcct.accountId;
  this.call(getFriendsListener, "youtube.users.list_friends", params);
};

YoutubeService.prototype._updateFriendAvatars =
function YoutubeService__updateFriendAvatars(aCoopAcct, aRefreshListener) {
  this._logger.debug("._updateFriendAvatars('" + aCoopAcct.id() + "', ...)");

  // Array of friends accounts (coop objects, not URNs) used to get avatars
  var friendsList = [];

  if (!this._updateFriendAvatarsComplete || aCoopAcct.isAuthenticated) {
    // Retrieve the account's friends list and add friends currently without an
    // avatar to a list to be updated at a later time
    var friendsEnum = aCoopAcct.friendsList.children.enumerate();
    while (friendsEnum.hasMoreElements()) {
      var friend = friendsEnum.getNext();
      if (friend.avatar == "") {
        this._logger.debug("adding friend with missing avatar for updating: "
                           + friend.accountId + "\n");
        friendsList.push(friend);
      }
    }
  }

  var inst = this;

  var friendsTimerCallback = {
    notify: function friendNotify(aTimer) {
      // Scrape the friend page for their avatar
      if (friendsList.length > 0 && aCoopAcct.isAuthenticated) {
        var friend = friendsList.shift();
        var page = gStrings["userprofile"]
                   .replace("%accountid%", friend.accountId);
        var xhr = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
                  .createInstance(CI.nsIXMLHttpRequest);
        xhr.open("GET", page, true);

        xhr.onreadystatechange = function friend_OnReadyStateChange(aEvent) {
          if (xhr.readyState == XMLHTTPREQUEST_READYSTATE_COMPLETED) {
            if (xhr.status == HTTP_CODE_OK) {
              var results = FlockSvcUtils.newResults();
              if (aCoopAcct.isAuthenticated &&
                  inst._WebDetective.detectNoDOM("youtube", "page", "",
                                                 xhr.responseText, results))
              {
                // If avatar returned as YT no avatar, then set coop avatar to
                // null and let people sidebar code set the Flock no avatar img.
                if (results.getPropertyAsAString("avatar")
                           .match(gStrings["noavatarimage"]))
                {
                  friend.avatar = null;
                  inst._logger.debug("no avatar for friend: " + friend.accountId + "\n");
                } else {
                  friend.avatar = results.getPropertyAsAString("avatar");
                  friend.avatar = inst._fixRelativeAvatarUrl(friend.avatar);
                  inst._logger.debug("found friend's avatar: " + friend.accountId
                                     + ", " + friend.avatar + "\n");
                }
              }
            }
          }
        };

        inst._logger.debug("requesting avatar for: " + page + "\n");
        xhr.send(null);

      } else {
        // Kill the timer if there's nothing left to update in the array or if
        // the account is no longer authenticated
        inst._logger.debug("no friends without avatars left to update, "
                           + "kill the timer");
        inst._friendsTimer.cancel();
        inst._friendsTimer = null;
        aRefreshListener.onSuccess(null, "_updateFriendAvatars");
      }
    }
  };

  // Start the timer to update the account's friends avatars
  if (!this._friendsTimer && friendsList.length > 0) {
    this._logger.debug("found friends without avatars to update, "
                       + "start the timer");
    this._friendsTimer = CC["@mozilla.org/timer;1"].createInstance(CI.nsITimer);
    this._friendsTimer.initWithCallback(friendsTimerCallback,
                                        FRIENDS_AVATAR_TIMEOUT,
                                        CI.nsITimer.TYPE_REPEATING_SLACK);

    // Once we've started updating friends without avatars, we only want to
    // check once per session so as not to incur a performance hit
    this._updateFriendAvatarsComplete = true;
  } else {
    aRefreshListener.onSuccess(null, "_updateFriendAvatars");
  }
};

YoutubeService.prototype._updateFriendRecentActivity =
function YoutubeService__updateFriendRecentActivity(aCoopAcct, aRefreshListener) {
  this._logger.debug("._updateFriendRecentActivity()");
  if (this._updateFriendActivityComplete ||
      this._friendActivityArray.length == 0 ||
      !aCoopAcct.isAuthenticated)
  {
    aRefreshListener.onSuccess(null, "_updateFriendRecentActivity");
    return;
  }
  var inst = this;

  var friendsTimerCallback = {
    notify: function friendNotify(aTimer) {
      inst._updateFriendActivityComplete = true;
      while (inst._friendActivityArray.length != 0) {
        var identity = inst._friendActivityArray.pop();
        inst._updateLastUpdateDate(identity);
        inst._logger.info("updating recent activity for " + identity.name);
      }
      inst._friendsActivityTimer.cancel();
      inst._updateFriendActivityComplete = false;
      inst._friendActivityArray = [];
      aRefreshListener.onSuccess(null, "_updateFriendRecentActivity");
      inst._logger.info("done updating recent activity. Killed the timer");
    }
  };

  // Start the timer to update the account's friends avatars
  this._logger.debug("found friends without avatars to update, "
                     + "start the timer");
  this._friendsActivityTimer = CC["@mozilla.org/timer;1"]
                               .createInstance(CI.nsITimer);
  this._friendsActivityTimer.initWithCallback(friendsTimerCallback,
                                              FRIENDS_ACTIVITY_TIMEOUT,
                                              CI.nsITimer.TYPE_REPEATING_SLACK);
  this._updateFriendActivityComplete = false;
};

YoutubeService.prototype._updateLastUpdateDate =
function YoutubeService__updateLastUpdateDate(aIdentity) {
  this._logger.info("._updateLastUpdateDate('"+ aIdentity.name + "')");
  
  var identity = aIdentity;
  var params = {
    user: identity.name,
    page: 1,
    per_page: 1
  };

  var inst = this;
  var myListener = {
    onResult: function _updateLastUpdateDate_onResult(aXML) {
      inst._logger.info("onResult");
      var video = inst._handlePhotosResult(aXML);
      if (video.length == 1) {
        video = video[0];
        //Youtube gives us UNIX time in seconds, change to miliseconds
        var time = video.uploadDate / 1000;
        identity.lastUpdate = time;
      }
    },
    onError: function _updateLastUpdateDate_onError(aError) {
      if (!identity.lastUpdate) {
        identity.lastUpdate = inst._numberFromString(identity.name);
      }
    }
  };

  this.call(myListener, "youtube.videos.list_by_user", params);
};


/**************************************************************************
 * YoutubeService: flockIRichContentDropHandler Implementation
 **************************************************************************/

/**
 * boolean handleDrop(in nsITransferable aFlavours,
 *                    in nsIDOMHTMLTextAreaElement aTextarea);
 * @see flockIRichContentDropHandler#handleDrop
 */
YoutubeService.prototype.handleDrop =
function YoutubeService_handleDrop(aFlavours, aTextarea) {
  this._logger.info(".handleDrop()");

  var dropCallback = function youtube_dropCallback(aFlav) {
    var data = {}, len = {};
    aFlavours.getTransferData(aFlav, data, len);
    var caretPos = aTextarea.selectionEnd;
    var currentValue = aTextarea.value;
    // Add a trailing space so that we don't mangle the url
    var nextChar = currentValue.charAt(caretPos);
    var trailingSpace = ((nextChar == "") ||
                         (nextChar == " ") || 
                         (nextChar == "\n"))
                      ? ""
                      : " ";
    // Only add a breadcrumb if the insertion point is at the end of
    // the text so that we don't duplicate breadcrumbs
    var breadcrumb = (aTextarea.value.length == aTextarea.selectionEnd)
                   ? CC[FLOCK_RDNDS_CONTRACTID]
                       .getService(CI.flockIRichDNDService)
                       .getBreadcrumb("plain")
                   : "";

    aTextarea.value = currentValue.substring(0, caretPos)
                    + data.value.QueryInterface(CI.nsISupportsString)
                          .data.replace(/: /, ":\n")
                    + trailingSpace
                    + currentValue.substring(caretPos)
                    + breadcrumb;
  };

  return this._handleTextareaDrop(CLASS_SHORT_NAME, this._c_svc.domains,
                                  aTextarea, dropCallback);
};

// DEFAULT: boolean _handleTextareaDrop(...)


/**************************************************************************
 * YoutubeService: nsIObserver Implementation
 **************************************************************************/

/**
 * void observe(nsISupports subject, char* topic, PRUnichar* data);
 * @see nsIObserver#observe
 */
YoutubeService.prototype.observe =
function YoutubeService_observe(aSubject, aTopic, aState) {
  switch (aTopic) {
    case "xpcom-shutdown":
      this._obs.removeObserver(this, "xpcom-shutdown");
      break;
  }
};


/**************************************************************************
 * YoutubeService: API methods
 **************************************************************************/

YoutubeService.prototype.call =
function YoutubeService_call(aListener, aMethod, aParams) {
  var url = "http://www.youtube.com/api2_rest?method=" + aMethod + "&dev_id=" + YT_DEVID;
  for (var p in aParams) {
    url += "&" + p + "=" + aParams[p];
  }
  var hr = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
           .createInstance(CI.nsIXMLHttpRequest)
           .QueryInterface(CI.nsIJSXMLHttpRequest);

  var inst = this;
  hr.onreadystatechange = function (aEvt) {
    if (hr.readyState == XMLHTTPREQUEST_READYSTATE_COMPLETED) {
      try {
        if (hr.status/100 == 2) {
          var rsp = hr.responseXML.getElementsByTagName("ut_response")[0];
          var stat = rsp.getAttribute("status");
          if (stat != "ok") {
            var error = inst.getError('SERVICE_ERROR', hr.responseXML, null);
            inst._logger.error(".call(..., '" + aMethod + "', ...) "
                               + "status is NOT OK; status = " + stat);
            aListener.onError(error);
          }
          else {
            aListener.onResult(hr.responseXML);
          }
        }
        else {
          // http errors
          inst._logger.error(".call(..., '" + aMethod + "', ...) "
                             + "HTTP Error; status = " + hr.status);
          aListener.onError(inst.getError("HTTP_ERROR", null, hr.status));
        }
      } catch(e) {
        // XMHTTPERROR (connection lost)
        inst._logger.error(".call(..., '" + aMethod + "', ...) "
                           + "XHR Error (connection lost); 9001\n" + e);
        aListener.onError(inst.getError("HTTP_ERROR", null, "9001"));
      }
    }
  };

  hr.mozBackgroundRequest = true;
  hr.open('GET',  url,true);
  this._logger.debug(".call(..., '" + aMethod + "', ...) url = " + url);
  hr.send(null);
};

YoutubeService.prototype.getError =
function YoutubeService_getError(aErrorType, aXML, aHTTPErrorCode) {
  var errorCode;
  var serviceErrorMessage;
  var error = CC["@flock.com/error;1"].createInstance(CI.flockIError);
  if (aErrorType == "HTTP_ERROR") {
    error.errorCode = aHTTPErrorCode;
  } else if (aErrorType == "SERVICE_ERROR") {
    try {
      errorCode = aXML.getElementsByTagName("error")[0].getAttribute('code');
      serviceErrorMessage = aXML.getElementsByTagName("error")[0].getAttribute('description');
    } catch (ex) {
      errorCode = "999"; // in case the error xml is invalid
    }

    switch (errorCode) { // http://www.youtube.com/dev_error_codes
      case "1":
        error.errorCode = error.HTTP_INTERNAL_SERVER_ERROR;
      break;

      case "2": // These errors are due to Flock sending a bad request
      case "3":
      case "4":
      case "5":
      case "6":
        error.errorCode = error.PHOTOSERVICE_INVALID_QUERY;
      break;

      case "7":
      case "8":
        error.errorCode = error.PHOTOSERVICE_INVALID_API_KEY;
        break;

      case "999":
        error.errorCode = error.PHOTOSERVICE_UNKNOWN_ERROR;
      break;

      default:
        error.errorCode = error.PHOTOSERVICE_UNKNOWN_ERROR;
      break;
    }
  }
  error.serviceErrorCode = errorCode;
  error.serviceErrorString = serviceErrorMessage;
  this._logger.error(error.errorString);
  return error;
};


/**************************************************************************
 * YoutubeService: flockISocialWebService Implementation
 **************************************************************************/

// readonly attribute long maxStatusLength;
YoutubeService.prototype.maxStatusLength = 0;

/**
 * void markAllMediaSeen(in AString aPersonUrn);
 * @see flockISocialWebService#markAllMediaSeen
 */
YoutubeService.prototype.markAllMediaSeen =
function YoutubeService_markAllMediaSeen(aIdentityUrn) {
  var identity = this._coop.get(aIdentityUrn);
  identity.unseenMedia = 0;
}; 


/**************************************************************************
 * YoutubeService: flockIWebService Implementation
 **************************************************************************/

// readonly attribute AString contractId;
// ALMOST duplicated by nsIClassInfo::contractID
// with different case: contractId != contractID
YoutubeService.prototype.contractId = CONTRACT_ID;

// readonly attribute AString icon;
YoutubeService.prototype.icon = YOUTUBE_FAVICON;

// readonly attribute AString shortName;
YoutubeService.prototype.shortName = CLASS_SHORT_NAME;

// readonly attribute AString title;
YoutubeService.prototype.title = CLASS_NAME;

// readonly attribute AString url;
YoutubeService.prototype.url = "http://www.youtube.com/";

// readonly attribute AString urn;
YoutubeService.prototype.urn = "urn:" + CLASS_SHORT_NAME + ":service";


/**************************************************************************
 * YoutubeService: flockIYoutubeService Implementation
 **************************************************************************/

/**
 * AString getVideoIDFromUrl(in AString aUrl);
 * @see flockIYoutubeService#getVideoIDFromUrl
 */
YoutubeService.prototype.getVideoIDFromUrl =
function youTubeService_getVideoIDFromUrl(aUrl) {
  var videoId = null;
  //http://www.youtube.com/v/lp_daXRkOH0
  //http://www.youtube.com/watch/v/YP2rgi978tw
  //http://www.youtube.com/watch?v=BjfbS_Kj-J0
  // /player2.swf?video_id=xPxDw7ajfGE&....
  var results = FlockSvcUtils.newResults();
  if (this._WebDetective.detectNoDOM("youtube",
                                     "videoId",
                                     "",
                                     aUrl,
                                     results))
  {
    videoId = results.getPropertyAsAString("videoId");
  }
  return videoId;
};


/*************************************************************************
 * Component: YoutubeAccount
 *************************************************************************/

function YoutubeAccount() {
  FlockSvcUtils.getLogger(this);
  this._logger.init("YoutubeAccount");

  FlockSvcUtils.getACUtils(this);
  FlockSvcUtils.getCoop(this);

  this._WebDetective = CC["@flock.com/web-detective;1"]
                       .getService(CI.flockIWebDetective);

  var sa = FlockSvcUtils.flockISocialAccount;
  sa.addDefaultMethod(this, "enumerateFriends");
  sa.addDefaultMethod(this, "getInviteFriendsURL");
  sa.addDefaultMethod(this, "formatFriendActivityForDisplay");
  sa.addDefaultMethod(this, "getFormattedFriendUpdateType");
  sa.addDefaultMethod(this, "getFriendCount");

  var wsa = FlockSvcUtils.flockIWebServiceAccount;
  wsa.addDefaultMethod(this, "getAllCustomParams");
  wsa.addDefaultMethod(this, "getCustomParam");
  wsa.addDefaultMethod(this, "getParam");
  wsa.addDefaultMethod(this, "getService");
  wsa.addDefaultMethod(this, "isAuthenticated");
  wsa.addDefaultMethod(this, "logout");
  wsa.addDefaultMethod(this, "setCustomParam");
  wsa.addDefaultMethod(this, "setParam");
}

YoutubeAccount.prototype = new FlockXPCOMUtils.genericComponent(
  CLASS_NAME + " Account",
  "",
  "",
  YoutubeAccount,
  0,
  [
    CI.flockIMediaAccount,
    CI.flockISocialAccount,
    CI.flockIWebServiceAccount,
    CI.flockIYoutubeAccount
  ]
);


/**************************************************************************
 * YoutubeAccount: flockISocialAccount Implementation
 **************************************************************************/

// readonly attribute boolean hasFriendActions;
YoutubeAccount.prototype.hasFriendActions = true;

// readonly attribute boolean isMyMediaFavoritesSupported;
YoutubeAccount.prototype.isMyMediaFavoritesSupported = true;

// readonly attribute boolean isPostLinkSupported;
YoutubeAccount.prototype.isPostLinkSupported = false;

// readonly attribute boolean isMeStatusSupported;
YoutubeAccount.prototype.isMeStatusSupported = false;

// readonly attribute boolean isFriendStatusSupported;
YoutubeAccount.prototype.isFriendStatusSupported = false;

// readonly attribute boolean isStatusEditable;
YoutubeAccount.prototype.isStatusEditable = false;

// DEFAULT: nsISimpleEnumerator enumerateFriends();

/**
 * AString formatStatusForDisplay(in AString aStatusMessage);
 * @see flockISocialAccount#formatStatusForDisplay
 */
YoutubeAccount.prototype.formatStatusForDisplay =
function YoutubeAccount_formatStatusForDisplay(aStatusMessage) {
  return aStatusMessage;
};

/**
 * AString getEditableStatus();
 * @see flockISocialAccount#getEditableStatus
 */
YoutubeAccount.prototype.getEditableStatus =
function YoutubeAccount_getEditableStatus() {
  return "";
};

/**
 * AString getFriendActions(in AString aFriendURN);
 * @see flockISocialAccount#getFriendActions
 */
YoutubeAccount.prototype.getFriendActions =
function YoutubeAccount_getFriendActions(aFriendURN) {
  this._logger.info(".getFriendActions('" + aFriendURN + "')");

  var actionNames = ["friendMessage",
                     "friendViewProfile",
                     "friendMediaFave",
                     "friendShareFlock"];

  var sbs = CC["@mozilla.org/intl/stringbundle;1"]
            .getService(CI.nsIStringBundleService);
  var bundle = sbs.createBundle(YOUTUBE_STRING_BUNDLE);

  var actions = [];
  var c_friend = this._coop.get(aFriendURN);
  if (c_friend) {
    var c_acct = this._coop.get(this.urn);
    for each (var i in actionNames) {
      actions.push({
        label: bundle.GetStringFromName("flock.youtube.actions." + i),
        class: i,
        spec: this._WebDetective.getString("youtube", i, "")
                  .replace("%accountid%", c_acct.accountId)
                  .replace("%friendid%", c_friend.accountId)
      });
    }
  }
  return JSON.toString(actions);
};

// DEFAULT: long getFriendCount();

/**
 * AString getMeNotifications();
 * @see flockISocialAccount#getMeNotifications
 */
YoutubeAccount.prototype.getMeNotifications =
function YoutubeAccount_getMeNotifications() {
  this._logger.info(".getMeNotifications()");

  var sbs = CC["@mozilla.org/intl/stringbundle;1"]
            .getService(CI.nsIStringBundleService);
  var bundle = sbs.createBundle(YOUTUBE_STRING_BUNDLE);

  var noties = [];
  var inst = this;
  function _addNotie(aType, aCount) {
    var stringName = "flock.youtube.noties."
                   + aType + "."
                   + ((parseInt(aCount) <= 0) ? "none" : "some");
    inst._logger.debug("aType " + aType
                       + " aCount " + aCount
                       + " name " + stringName);
    noties.push({
      class: aType,
      tooltip: bundle.GetStringFromName(stringName),
      metricsName: aType,
      count: aCount,
      URL: inst._WebDetective.getString("youtube", aType + "_URL", "")
    });
  }
  var c_acct = this._coop.get(this.urn);
  _addNotie("meMessages", c_acct.accountMessages);
  return JSON.toString(noties);
};

/**
 * AString getPostLinkAction(in nsITransferable aTransferable);
 * @see flockISocialAccount#getPostLinkAction
 */
YoutubeAccount.prototype.getPostLinkAction =
function YoutubeAccount_getPostLinkAction(aTransferable) {
  return "";
};

/**
 * AString getProfileURLForFriend(in AString aFriendURN);
 * @see flockISocialAccount#getProfileURLForFriend
 */
YoutubeAccount.prototype.getProfileURLForFriend =
function YoutubeAccount_getProfileURLForFriend(aFriendURN) {
  this._logger.debug(".getProfileURLForFriend('" + aFriendURN + "')");

  var url = "";
  var c_friend = this._coop.get(aFriendURN);
  if (c_friend) {
    url = this._WebDetective.getString("youtube", "friendprofile", "")
                            .replace("%accountid%", c_friend.accountId);
  }
  return url;
};

/**
 * AString getSharingAction(in AString aFriendURN,
 *                          in nsITransferable aTransferable);
 * @see flockISocialAccount#getSharingAction
 */
YoutubeAccount.prototype.getSharingAction =
function YoutubeAccount_getSharingAction(aFriendURN, aTransferable) {
  this._logger.info(".getSharingAction('" + aFriendURN + "')");

  var sharingAction = "";
  var c_friend = this._coop.get(aFriendURN);
  if (c_friend) {
    var flavors = ["text/x-flock-media",
                   "text/x-moz-url",
                   "text/unicode",
                   "text/html"];

    var message = CC[FLOCK_RDNDS_CONTRACTID]
                  .getService(CI.flockIRichDNDService)
                  .getMessageFromTransferable(aTransferable,
                                              flavors.length,
                                              flavors);
    if (!message.body) {
      return sharingAction;
    } else {
      var videoId = this.getService()
                        .QueryInterface(CI.flockIYoutubeService)
                        .getVideoIDFromUrl(message.body);
      if (videoId) {
        sharingAction = this._WebDetective.getString("youtube",
                                                     "shareAction_sharevideo",
                                                     "");
      } else {
        sharingAction = this._WebDetective.getString("youtube",
                                                     "shareAction_privatemessage",
                                                     "");
      }
    }
  }
  this._logger.info(sharingAction);
  return sharingAction;
};

/**
 * void markAllMeNotificationsSeen(in AString aType);
 * @see flockISocialAccount#markAllMeNotificationsSeen
 */
YoutubeAccount.prototype.markAllMeNotificationsSeen =
function YoutubeAccount_markAllMeNotificationsSeen(aType) {
  this._logger.debug(".markAllMeNotificationsSeen('" + aType + "')");
  var c_acct = this._coop.get(this.urn);
  switch (aType) {
    case "meMessages":
      c_acct.accountMessages = 0;
      break;
    default:
      break;
  }
};

/**
 * void setStatus(in AString aStatusMessage, in flockIListener aFlockListener);
 * @see flockISocialAccount#setStatus
 */
YoutubeAccount.prototype.setStatus =
function YoutubeAccount_setStatus(aStatusMessage, aFlockListener) {
};


/**************************************************************************
 * YoutubeAccount: flockIWebServiceAccount Implementation
 **************************************************************************/

// readonly attribute AString urn;
YoutubeAccount.prototype.urn = "";

// DEFAULT: nsIPropertyBag getAllCustomParams();
// DEFAULT: nsIVariant getCustomParam(in AString aParamName);
// DEFAULT: nsIVariant getParam(in AString aParamName);
// DEFAULT: flockILoginWebService getService();
// DEFAULT: boolean isAuthenticated();

/**
 * void keep();
 * @see flockIWebServiceAccount#keep
 */
YoutubeAccount.prototype.keep =
function YoutubeAccount_keep() {
  this._logger.debug(".keep()");
  var c_acct = this._coop.get(this.urn);
  c_acct.isTransient = false;
  var key = this.getService().urn + ":" + c_acct.accountId;
  this._acUtils.makeTempPasswordPermanent(key);
};

/**
 * void login(in flockIListener aFlockListener);
 * @see flockIWebServiceAccount#login
 */
YoutubeAccount.prototype.login =
function YoutubeAccount_login(aFlockListener) {
  this._logger.debug(".login(...)");
  this._acUtils.ensureOnlyAuthenticatedAccount(this.urn);
  this.setParam("isPollable", true);
  // Force refresh on login
  var pollerSvc = CC["@flock.com/poller-service;1"]
                  .getService(CI.flockIPollerService);
  pollerSvc.forceRefresh(this.urn);
  if (aFlockListener) {
    aFlockListener.onSuccess();
  }
};

// DEFAULT: void logout(in flockIListener aFlockListener);
// DEFAULT: void setCustomParam(in AString aParamName, in nsIVariant aValue);
// DEFAULT: void setParam(in AString aParamName, in nsIVariant aValue);


/**************************************************************************
 * YoutubeAccount: flockIYoutubeAccount Implementation
 **************************************************************************/

/**
 * void shareFlock(in AString aFriendURN);
 * @see flockIYoutubeAccount#shareFlock
 */
YoutubeAccount.prototype.shareFlock =
function YoutubeAccount_shareFlock(aFriendURN) {
  this._logger.info(".shareFlock('" + aFriendURN + "')");
  
  var sbs = CC["@mozilla.org/intl/stringbundle;1"]
              .getService(CI.nsIStringBundleService);
  var bundle = sbs.createBundle(YOUTUBE_STRING_BUNDLE);
  var body = bundle.GetStringFromName("flock.youtube.friendShareFlock.message");
  var subj = bundle.GetStringFromName("flock.youtube.friendShareFlock.subject");
  this._composeMessage(aFriendURN, subj, body, false);
};

/**
 * void youtubePrivateMessage(in AString aFriendURN,
 *                            in nsITransferable aTransferable);
 * @see flockIYoutubeAccount#youtubePrivateMessage
 */
YoutubeAccount.prototype.youtubePrivateMessage =
function YoutubeAccount_youtubePrivateMessage(aFriendURN, aTransferable) {
  this._logger.info(".youtubePrivateMessage('" + aFriendURN + "', ...)");

  var flavors = ["text/x-flock-media",
                 "text/x-moz-url",
                 "text/unicode",
                 "text/html"];

  var message = CC[FLOCK_RDNDS_CONTRACTID]
                .getService(CI.flockIRichDNDService)
                .getMessageFromTransferable(aTransferable,
                                            flavors.length,
                                            flavors);
  if (message.body) {
    this._composeMessage(aFriendURN, message.subject, message.body, true);
  }
};

/**
 * void youtubeShareVideo(in AString aFriendURN,
 *                        in nsITransferable aTransferable);
 * @see flockIYoutubeAccount#youtubeShareVideo
 */
YoutubeAccount.prototype.youtubeShareVideo =
function YoutubeAccount_youtubeShareVideo(aFriendURN, aTransferable) {
  this._logger.info(".youtubeShareVideo('" + aFriendURN + "', ...)");

  var flavors = ["text/x-flock-media",
                 "text/x-moz-url",
                 "text/unicode",
                 "text/html"];

  var message = CC[FLOCK_RDNDS_CONTRACTID]
                .getService(CI.flockIRichDNDService)
                .getMessageFromTransferable(aTransferable,
                                            flavors.length,
                                            flavors);
  if (message.body) {
    var inst = this;
    var share_url = inst._WebDetective.getString("youtube",
                                                 "shareIframeURL", "");
    var obs = CC["@mozilla.org/observer-service;1"]
                .getService(CI.nsIObserverService);
    var videoId = this.getService()
                      .QueryInterface(CI.flockIYoutubeService)
                      .getVideoIDFromUrl(message.body);
    var url = this._WebDetective.getString("youtube", "youtubeShare_URL", "")
                  .replace("%videoid%", videoId);
    var wm = CC["@mozilla.org/appshell/window-mediator;1"]
             .getService(CI.nsIWindowMediator);
    var win = wm.getMostRecentWindow("navigator:browser");
    var dimen = inst._WebDetective.getString("youtube",
                                             "shareAction_sharevideo_features",
                                             "");
    win.open(url, "Share", dimen);
    win = wm.getMostRecentWindow("navigator:browser");
    var flockDocumentReadyObserver = {
      observe: function openShareVideo_FormFill(aSubject, aTopic, aData) {
        if (aData == share_url) {
          obs.removeObserver(this, "FlockDocumentReady");
          var contentWindow = win.gBrowser.docShell
                                 .QueryInterface(CI.nsIInterfaceRequestor)
                                 .getInterface(CI.nsIDOMWindow);

          function insertContent(aDoc, aXpathquery, aAction, aMessage) {
            var formItems = aDoc.evaluate(aXpathquery, aDoc, null,
                                          CI.nsIDOMXPathResult.ANY_TYPE, null);
            if (formItems) {
              var formItem = formItems.iterateNext();
              if (!formItem) {
                return;
              }
              switch (aAction) {
                case "click":
                  formItem.click();
                  break;

                case "username":
                  if (formItem.hasAttribute("value")) {
                    formItem.setAttribute("value", aMessage);
                  }
                  break;

                case "message":
                  var textNode = doc.createTextNode(aMessage);
                  formItem.appendChild(textNode);
                  break;
              
                default:
                  inst._logger.debug("no action associated with " + aAction);
              }
            }
          }
          var doc = contentWindow.document;
          // Insert Flock bread crumb
          var breadcrumb = CC[FLOCK_RDNDS_CONTRACTID]
                             .getService(CI.flockIRichDNDService)
                             .getBreadcrumb("plain");
          var webDicName = "";
          var xpathquery = "";
          if (breadcrumb) {
            webDicName = "youtubeshareVideo_message";
            xpathquery = inst._WebDetective.getString("youtube",
                                                      webDicName, "");
            insertContent(doc, xpathquery, "message", breadcrumb);
          }

          // Insert friend trying to save to into iframe
          webDicName = "youtubeshareVideo_IframeXPath";
          xpathquery = inst._WebDetective.getString("youtube",
                                                   webDicName, "");
          var formItems = doc.evaluate(xpathquery, doc, null,
                                       CI.nsIDOMXPathResult.ANY_TYPE, null);
          if (formItems) {
            var iframe = formItems.iterateNext();
            if (iframe) {
              var c_friend = inst._coop.get(aFriendURN);
              webDicName = "youtubeshareVideo_friendCheckXPath";
              xpathquery = inst._WebDetective.getString("youtube", webDicName, "")
                                             .replace("%friendname%", c_friend.name);
              insertContent(iframe.contentDocument, xpathquery, "click");
            }
          }
        }
      }
    };
    obs.addObserver(flockDocumentReadyObserver, 'FlockDocumentReady', false);
  }
};

// account helper
YoutubeAccount.prototype._addMediaToFavorites =
function YoutubeAccount__addMediaToFavorites(aAccountId) {
  var sbs = CC["@mozilla.org/intl/stringbundle;1"]
            .getService(CI.nsIStringBundleService);
  var bundle = sbs.createBundle(YOUTUBE_STRING_BUNDLE);
  
  var svc = this.getService();

  var id = FAV_CHANNELS_ROOT + ":" + svc.shortName
         + ":favorites:" + aAccountId;
  var name = bundle.formatStringFromName("flock.youtube.favoritesLC",
                                         [aAccountId], 1);
  var item = new this._coop.MediaQuery(id, {
    name: name,
    query: "favorites:" + aAccountId,
    service: svc.shortName,
    favicon: svc.icon,
    isPollable: true,
    hasUnseenItems: false
  });

  this._coop.media_favorites.children.addOnce(item);
};

YoutubeAccount.prototype._composeMessage =
function YoutubeAccount__composeMessage(aFriendURN,
                                        aSubject,
                                        aBody,
                                        addBreadCrumb)
{
  var body = aBody;
  var subject = aSubject;
  var c_friend = this._coop.get(aFriendURN);
  var url = this._WebDetective.getString("youtube", "youtubeMessage_URL", "")
                .replace("%friendid%", c_friend.accountId);
  var wm = CC["@mozilla.org/appshell/window-mediator;1"]
             .getService(CI.nsIWindowMediator);
  var win = wm.getMostRecentWindow("navigator:browser");
  if (win) {
   var browser = win.getBrowser();
   var newTab = browser.loadOneTab(url, null, null, null, false, false);
   var obs = CC["@mozilla.org/observer-service;1"]
             .getService(CI.nsIObserverService);
   var inst = this;
   var observer = {
     observe: function openSendMessageTabForFill_observer(aContent,
                                                          aTopic,
                                                          aContextUrl)
     {
       var contentWindow = newTab.linkedBrowser.docShell
                                 .QueryInterface(CI.nsIInterfaceRequestor)
                                 .getInterface(CI.nsIDOMWindow);
       function insertContent(aWebDicString, aMessage) {
         var xpathquery = inst._WebDetective.getString("youtube", aWebDicString, "");
         var doc = contentWindow.document;
         var formItems = doc.evaluate(xpathquery, doc, null,
                                      CI.nsIDOMXPathResult.ANY_TYPE, null);
         if (formItems) {
           var formItem = formItems.iterateNext();
           formItem.value = aMessage;
         }
       }
       if (contentWindow == aContent) {
         obs.removeObserver(this, "EndDocumentLoad");
         insertContent("youtubemessage_subjectXPath", subject);
         if (addBreadCrumb) {
           // Add breadcrumb to message body
            var breadcrumb = CC[FLOCK_RDNDS_CONTRACTID]
                               .getService(CI.flockIRichDNDService)
                               .getBreadcrumb("plain");
           if (breadcrumb) {
             body += breadcrumb;
           }
        }
         insertContent("youtubemessage_bodyXPath", body);
       }
     }
   };
   obs.addObserver(observer, "EndDocumentLoad", false);
  }
};


/*************************************************************************
 * XPCOM Support - Module Construction
 *************************************************************************/

// Create array of components.
var componentsArray = [YoutubeService];

// Generate a module for XPCOM to find.
var NSGetModule = FlockXPCOMUtils.generateNSGetModule(MODULE_NAME,
                                                      componentsArray);
