// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

var EXPORTED_SYMBOLS = ["FlockLiveSearchProvider"];

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

CU.import("resource:///modules/FlockXPCOMUtils.jsm");
CU.import("resource:///modules/FlockStringBundleHelpers.jsm");


const NAME_SUFFIX = "Live Search";

var FlockLiveSearchProvider = {
  generateComponent: function FLSP_generateComponent(aName,
                                                     aCID,
                                                     aCtor,
                                                     aIconURL,
                                                     aFullResultsURL)
  {
    var shortName = aName.replace(/\s+/, "").toLowerCase();
    var className = aName + " " + NAME_SUFFIX;
    var contractID = "@flock.com/live-search/" + shortName + ";1";

    function addDefaultIcons(aResults) {
      for each (let result in aResults) {
        if (!result.icon) {
          result.icon = aIconURL;
        }
      }
    }

    var prototype = new FlockXPCOMUtils.genericComponent(
      className, aCID, contractID, aCtor,
      CI.nsIClassInfo.SINGLETON, [CI.flockISearchService]
    );

    prototype._xpcom_categories = [
      { category: "flockISearchService", entry: shortName }
    ];

    prototype._init = function FLSP__init() {
      this._logger = CC["@flock.com/logger;1"].createInstance(CI.flockILogger);
      this._logger.init(shortName + "livesearch");
    };

    prototype.__defineGetter__("serviceName",
      function FLSP_serviceName() {
        return flockGetString("search/search",
                              "flock.search.api." + shortName);
      }
    );

    prototype.__defineGetter__("shortName", function FLSP_shortName() {
      return shortName;
    });

    prototype.__defineGetter__("icon", function FLSP_icon() {
      return aIconURL;
    });

    prototype.__defineGetter__("fullResultsUrl", function FLSP_fullResultsUrl() {
      return aFullResultsURL;
    });

    prototype.search = function FLSP_search(aQuery,
                                            aNumResults,
                                            aSearchListener)
    {
      var inst = this;

      if (!aQuery.length) {
        return;
      }

      var url = inst._getQueryURL(aQuery, aNumResults);
      this._logger.debug("call: " + url);

      var xhr = CC["@mozilla.org/xmlextras/xmlhttprequest;1"]
                .createInstance(CI.nsIXMLHttpRequest);

      xhr.onload = function FLSP_onload(aEvent) {
        if (xhr.status == 200 || xhr.status == 201) {
          var results;
          try {
            results = inst._getResults(xhr.responseXML, aNumResults);
          } catch (ex) {
            inst._logger.error("getResults failed: " + ex);
            results = [];
          }

          addDefaultIcons(results);
          aSearchListener.foundResults(results, shortName, aQuery);
        }
      };

      xhr.mozBackgroundRequest = true;
      xhr.open("GET", url);
      xhr.send(null);
    };

    return prototype;
  },

  generateNSGetModule: function FLSP_generateNSGetModule(aName, aCtor) {
    return FlockXPCOMUtils.generateNSGetModule(aName + " " + NAME_SUFFIX,
                                               [aCtor]);
  }
};
